/* RTOS low-level code */
/* Carousel variant */

#define FROM_ASM
#include "rtos.h"


BIW_offset = BIW - switch_area
BIW_p_offset = BIW + 3 - switch_area
BIW_stack_size_offset = BIW + 2 - switch_area
BIW_code_size_offset = BIW + 1 - switch_area
T_offset = T - switch_area

rBIW_offset = 0x00
r1_offset = 0x04
r2_offset = 0x08
r3_offset = 0x0c
r4_offset = 0x10
r5_offset = 0x14
r6_offset = 0x18
r7_offset = 0x1c
r8_offset = 0x20
r9_offset = 0x24
r10_offset = 0x28
r11_offset = 0x2c
r12_offset = 0x30
r13_offset = 0x34
r14_offset = 0x38
r15_offset = 0x3c
r15_b_offset = 0x40
r18_offset = 0x44
r19_offset = 0x48
r1_b_offset = 0x4c
rmsr_offset = 0x50
sys_blocks_offset = 0x54
saved_block_bases = 0x58
switch_area_size = saved_block_bases + (4 * NUM_BLOCKS)
current_time = TIMER_REGISTER


# defined here
.global reset
.global call_function
.global exit
.global _exit
.global memmove
.global memcpy
.global _open
.global print
.global notify_task_name
.global current_time
.global kernel_idle

# defined elsewhere
.global scheduler
.global main

	.section	.kernelstuff.t,"ax",@progbits
.org 0x0
reset:
        # SYSTEM mode: interrupts off, breaks on
        msrset  r3, MSR_BE
        bri     kernel_entry
.org 0x8
exception:
        brai    exception       # crash
.org 0x10
interrupt:
        brai    interrupt_handler
.org 0x18
break:
        brai    break           # crash
.org 0x20
hw_exception:
        brai    hw_exception    # crash

.org 0x50

switch_area:
        .space  switch_area_size
sched_stack_end:

T:
        .byte   0
.align 2
BIW:
        .long   0
.align 2
switch_stack_ptr:
        .long   0

kernel_entry:
        # Kernel entry point, set up globals
        addik   r3, r0, SWITCH_STACK
        swi     r3, r0, switch_stack_ptr
        addik   r1, r0, SYSTEM_STACK
        addik   r17, r0, switch_area

        # Initialise C runtime
        addi	r6, r0, __bss_start             /* clear BSS */
        addi	r7, r0, __bss_end		
        rsub	r18, r6, r7		
        blei	r18, .Lendbss
.Lloopbss:	
        swi	    r0, r6, 0
        addi	r6, r6, 4
        rsub	r18, r6, r7
        bgti	r18, .Lloopbss
.Lendbss:

        addi    r5, r0, 0
        addi    r6, r0, 0

        brlid	r15, main
        addi    r7, r0, 0

_exit:
exit:   # SYSTEM mode: disable interrupts
        msrclr  r3, MSR_IE
        addik   r3, r0, exit_called
        cput    r3, rfsl2
0:      .long   0x80000001
        bri     0b          # exit loop
       
task_func_is_zero:
        addik   r5, r0, task_func_is_zero_called
        cput    r5, rfsl2
        bri     exit
    
interrupt_handler:
        # Save registers not saved by call_function
        swi     r3, r17, r3_offset
        swi     r4, r17, r4_offset
        swi     r5, r17, r5_offset
        swi     r6, r17, r6_offset
        swi     r7, r17, r7_offset
        swi     r8, r17, r8_offset
        swi     r9, r17, r9_offset
        swi     r10,r17, r10_offset
        swi     r11,r17, r11_offset
        swi     r12,r17, r12_offset
        swi     r14,r17, r14_offset
        swi     r15,r17, r15_b_offset
        swi     r18,r17, r18_offset
        swi     r19,r17, r19_offset
        swi     r1,r17, r1_b_offset

        # New stack just for this function
        addi    r1, r0, sched_stack_end

        # Where shall we call next?
        brlid   r15, scheduler
        nop

        lwi     r1, r17, r1_b_offset
        or      r19, r3, r0

        beqid   r3, do_not_task_switch
        or      r5, r0, r0

        lwi     r5, r3, 0       # get task function

        # Notify: a new task will begin
        beqid   r5, task_func_is_zero
        addik   r4, r0, TRC_TASK_START
        put     r5, rfsl0
        cput    r4, rfsl4 

        # Call that task
        brlid   r15, call_function
        lwi     r6, r3, 8       # get block info word

        swi     r0, r19, 4      # clear running flag
        lwi     r5, r19, 0      # get task function

        # Preempt immediately on return from ISR in case we have 
        # another task waiting to run
        swi     r0, r0, current_time

        # Notify: task finished
        addik   r4, r0, TRC_TASK_END
        put     r5, rfsl0
        cput    r4, rfsl4 

do_not_task_switch:
        # Notify: returning from interrupt
        addik   r4, r0, TRC_IN_RTI
        put     r5, rfsl0
        cput    r4, rfsl4 

        lwi     r3, r17, r3_offset
        lwi     r4, r17, r4_offset
        lwi     r5, r17, r5_offset
        lwi     r6, r17, r6_offset
        lwi     r7, r17, r7_offset
        lwi     r8, r17, r8_offset
        lwi     r9, r17, r9_offset
        lwi     r10,r17, r10_offset
        lwi     r11,r17, r11_offset
        lwi     r12,r17, r12_offset
        lwi     r14,r17, r14_offset
        lwi     r15,r17, r15_b_offset
        lwi     r18,r17, r18_offset

        # Return from interrupt
        rtid    r14, 0
        lwi     r19,r17, r19_offset


call_function:
        # r5: address of target function
        # r6: block info word
        # r7-r10: parameters of target function
        # r15: return PC - 8

        # SYSCALL mode: disable interrupts, breaks
        msrclr  r3, MSR_IE | MSR_BE

        # Old BIW (to be saved)
        lwi     r4, r17, BIW_offset

        # Save some registers
        swi     r1, r17, r1_offset
        swi     r15,r17, r15_offset
        swi     r3, r17, rmsr_offset
        swi     r4, r17, rBIW_offset

        # New block info word stored at BIW
        swi     r6, r17, BIW_offset

        # Load T and then p
        lbui    r3, r17, T_offset
        lbui    r4, r17, BIW_p_offset

swap_out:      
        # Swap out Carousel blocks
        addik   r4, r4, -1  # p := p - 1

        # Save the base register
        bslli   r11, r3, 2
        lwi     r16, r11, BASE_REGISTER_AREA
        swi     r16, r11, saved_block_bases + switch_area

        # CLOSE(r3)
        addi    r11, r0, -1
        put     r11, rfsl0      # src (-1 for close)
        put     r3, rfsl1       # dest as block
        cput    r0, rfsl1       # 1 block only

        addik   r3, r3, 1       # T := T + 1
        bneid   r4, swap_out
        andi    r3, r3, NUM_BLOCKS - 1

        # Switching area is pushed onto the switch_stack
        addi    r4, r0, switch_stack_ptr
        lw      r16, r0, r4
        addik   r11, r0, switch_area_size
        rsub    r16, r11, r16

        # DMA(r17 to r16)
        put     r17, rfsl0       # src
        put     r16, rfsl1       # dest
        cput    r11, rfsl0       # size

        sw      r16, r0, r4
        
        # Update T
        sbi     r3, r17, T_offset

        # How many stack blocks required for function?
        lbui    r12, r17, BIW_stack_size_offset
        beqid   r12, no_stack_required
        rsub    r11, r12, r3            # T subtracted by stack blocks

        andni   r1, r1, BLOCK_SIZE - 1  # round r1 down
        bslli   r4, r12, BLOCK_SIZE_SHIFT
        rsub    r4, r4, r1              # stack space allocated

        andi    r11, r11, NUM_BLOCKS - 1
        ori     r12, r12, OPEN_SCRATCH_FLAG

        # OPEN(r4, r11, r12)
        put     r4, rfsl0       # src
        put     r11, rfsl1      # dest as block
        cput    r12, rfsl1      # size (block mode)

no_stack_required:
        # How many code blocks required for function?
        lbui    r12, r17, BIW_code_size_offset
        rsub    r11, r12, r11           # T subtracted by code blocks
        andi    r11, r11, NUM_BLOCKS - 1
        ori     r12, r12, OPEN_DEFER_FLAG
        
        # OPEN(r5, r11, r12)
        put     r5, rfsl0       # src
        put     r11, rfsl1      # dest as block
        cput    r12, rfsl1      # size (block mode)

        # USER mode: enable interrupts, breaks
        msrset  r3, MSR_IE | MSR_BE

        # Parameter registers moved into place
        or      r4, r7, r0
        or      r6, r8, r0
        or      r7, r9, r0
        or      r8, r10, r0

        # Call function
        brald   r15, r5
        or      r5, r4, r0

return_point:
        # SYSCALL mode: disable interrupts, breaks
        msrclr  r5, MSR_IE | MSR_BE

        # Switching area is popped off the switch_stack
        addik   r17, r0, switch_area

        addi    r5, r0, switch_stack_ptr
        addik   r11, r0, switch_area_size
        lw      r16, r0, r5

        # DMA(r16 to r17)
        put     r16, rfsl0       # src
        put     r17, rfsl1       # dest
        cput    r11, rfsl0       # size

        add     r16, r16, r11
        sw      r16, r0, r5

        # What is p?
        lbui    r5, r17, BIW_p_offset
        lbui    r6, r17, T_offset

        # Discard all blocks used by the system
        rsub    r11, r5, r6     # T - p
        andi    r11, r11, NUM_BLOCKS - 1
        ori     r12, r5, OPEN_DISCARD_FLAG
        addi    r7, r0, -1      # CLOSE: src = -1
        put     r7, rfsl0       # src
        put     r11, rfsl1      # dest as block
        cput    r12, rfsl1       # size (block mode)

        # Swap in Carousel blocks
swap_in:      
        addik   r5, r5, -1  # p := p - 1
        addik   r6, r6, -1  # T := T - 1
        andi    r6, r6, NUM_BLOCKS - 1

        # Restore the base register
        bslli   r11, r6, 2
        lwi     r16, r11, saved_block_bases + switch_area

        # OPEN(r16, r6)
        put     r16, rfsl0      # src
        put     r6, rfsl1       # block number
        cput    r0, rfsl1       # 1 block only

        bnei    r5, swap_in

        # Update T
        sbi     r6, r0, T

        # Restore context
        lwi     r1, r17, r1_offset
        lwi     r15,r17, r15_offset
        lwi     r16,r17, rmsr_offset
        lwi     r5, r17, rBIW_offset
        swi     r5, r17, BIW_offset

        # return
        rtsd    r15, 8
        # Entry status returned (enable interrupts if
        # returning to user code)
        mts     rmsr, r16

memcpy:
memmove: # r5 is dest
        # r6 is source
        # r7 is size

        # SYSCALL mode: disable interrupts, breaks
        msrclr  r3, MSR_IE | MSR_BE

        # DMA(r6 to r5)
        put     r6, rfsl0        # src
        put     r5, rfsl1        # dest
        cput    r7, rfsl0        # size
        or      r3, r5, r0

        rtsd    r15, 8
        # USER mode: enable interrupts, breaks
        msrset  r5, MSR_IE | MSR_BE

_open:  # r5 is base address
        # r6 is block number
        # r7 is number of blocks

        # SYSCALL mode: disable interrupts, breaks
        msrclr  r3, MSR_IE | MSR_BE

        # r6 is relativised to T-p
        lbui    r3, r17, T_offset
        addk    r3, r3, r6                      # T + block
        lbui    r4, r17, BIW_p_offset
        rsub    r3, r4, r3                      # T + block - p
        andi    r6, r3, NUM_BLOCKS - 1

        put     r5, rfsl0       # src
        put     r6, rfsl1       # dest as block
        cput    r7, rfsl1       

        rtsd    r15, 8
        # USER mode: enable interrupts, breaks
        msrset  r5, MSR_IE | MSR_BE
       
print:  # r5 is address of 0-terminated string
        rtsd    r15, 8
        cput    r5, rfsl2

notify_task_name:
        # r5 = task name
        # r6 = function ptr
        put     r6, rfsl0       # function
        rtsd    r15, 8
        cput    r5, rfsl5       # task name

kernel_idle:
        # Allocate all Carousel blocks to null space
        addik   r5, r0, EMPTY_SPACE_AREA
        addik   r6, r0, NUM_BLOCKS
        put     r5, rfsl0       # src
        put     r0, rfsl1      # dest as block
        cput    r6, rfsl1      # size (block mode)

        # Notify: idle task start
        addik   r5, r0, TRC_TASK_START
        put     r0, rfsl0
        cput    r5, rfsl4 

        # Clock start - preempt as soon as interrupts are enabled
        # The clock has to be enabled in SYSTEM mode for the sake of the 
        # simulator which assumes that stopped clock == SYSTEM.
        swi     r0, r0, current_time

        # USER mode: enable interrupts, breaks
        msrset  r5, MSR_IE | MSR_BE

        # Idle loop
0:      addik   r5, r0, idle_called
        cput    r5, rfsl2 # print
        cput    r5, rfsl3 # HALT
        bri     0b

.data
exit_called:
        .ascii  "exit called"
        .byte   10
        .byte   0
idle_called:
        .ascii  "idle"
        .byte   10
        .byte   0
task_func_is_zero_called:
        .ascii  "Task function pointer is zero!"
        .byte   10
        .byte   0


